/*****************************************************************************
 * Copyright by The HDF Group.                                               *
 * Copyright by the Board of Trustees of the University of Illinois.         *
 * All rights reserved.                                                      *
 *                                                                           *
 * This file is part of the HDF Java Products distribution.                  *
 * The full copyright notice, including terms governing use, modification,   *
 * and redistribution, is contained in the file, COPYING.                    *
 * COPYING can be found at the root of the source code distribution tree.    *
 * If you do not have access to this file, you may request a copy from       *
 * help@hdfgroup.org.                                                        *
 ****************************************************************************/

/*
 *  For details of the HDF libraries, see the HDF Documentation at:
 *    http://hdfgroup.org/HDF5/doc/
 *
 */

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#include <jni.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "hdf5.h"
#include "h5util.h"
#include "h5gImp.h"

/*
 * Pointer to the JNI's Virtual Machine; used for callback functions.
 */
/* extern JavaVM *jvm; */

/*
 * Class:     hdf_hdf5lib_H5
 * Method:    _H5Gclose
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL
Java_hdf_hdf5lib_H5__1H5Gclose
    (JNIEnv *env, jclass clss, jint group_id)
{
    herr_t retVal = FAIL;

    UNUSED(clss);

    if ((retVal = H5Gclose((hid_t)group_id)) < 0)
        H5_LIBRARY_ERROR(ENVONLY);

done:
    return (jint)retVal;
} /* end Java_hdf_hdf5lib_H5__1H5Gclose */

/*
 * Create a java object of hdf.h5.structs.H5G_info_t
 * public class H5G_info_t {
 *   public H5G_STORAGE_TYPE  storage_type; // Type of storage for links in group
 *   public long     nlinks;       // Number of links in group
 *   public long     max_corder;   // Current max. creation order value for group
 *   public int      mounted;      // Whether group has a file mounted on it
 * }
 *
 */
jobject
create_H5G_info_t
    (JNIEnv *env, H5G_info_t group_info)
{
    jfieldID fid_storage_type, fid_nlinks, fid_max_corder, fid_mounted;
    jboolean jmounted;
    jclass   cls;
    jint     storage_type;
    jobject  obj = NULL;

    if (NULL == (cls = ENVPTR->FindClass(ENVONLY, "hdf/hdf5lib/structs/H5G_info_t")))
        CHECK_JNI_EXCEPTION(ENVONLY, JNI_FALSE);

    if (NULL == (obj = ENVPTR->AllocObject(ENVONLY, cls)))
        CHECK_JNI_EXCEPTION(ENVONLY, JNI_FALSE);

    if (NULL == (fid_storage_type = ENVPTR->GetFieldID(ENVONLY, cls, "storage_type", "I")))
        CHECK_JNI_EXCEPTION(ENVONLY, JNI_FALSE);

    if (NULL == (fid_nlinks = ENVPTR->GetFieldID(ENVONLY, cls, "nlinks", "J")))
        CHECK_JNI_EXCEPTION(ENVONLY, JNI_FALSE);

    if (NULL == (fid_max_corder = ENVPTR->GetFieldID(ENVONLY, cls, "max_corder", "J")))
        CHECK_JNI_EXCEPTION(ENVONLY, JNI_FALSE);

    if (NULL == (fid_mounted = ENVPTR->GetFieldID(ENVONLY, cls, "mounted", "Z")))
        CHECK_JNI_EXCEPTION(ENVONLY, JNI_FALSE);

    jmounted = (group_info.mounted == 0) ? JNI_FALSE : JNI_TRUE;
    storage_type = (jint)group_info.storage_type;

    ENVPTR->SetIntField(ENVONLY, obj, fid_storage_type, (jint)storage_type);
    CHECK_JNI_EXCEPTION(ENVONLY, JNI_FALSE);

    ENVPTR->SetLongField(ENVONLY, obj, fid_nlinks, (jlong)group_info.nlinks);
    CHECK_JNI_EXCEPTION(ENVONLY, JNI_FALSE);

    ENVPTR->SetLongField(ENVONLY, obj, fid_max_corder, (jlong)group_info.max_corder);
    CHECK_JNI_EXCEPTION(ENVONLY, JNI_FALSE);

    ENVPTR->SetBooleanField(ENVONLY, obj, fid_mounted, jmounted);
    CHECK_JNI_EXCEPTION(ENVONLY, JNI_FALSE);

done:
    return obj;
} /* end create_H5G_info_t */

JNIEXPORT jint JNICALL Java_hdf_hdf5lib_H5__1H5Gcreate
    (JNIEnv *env, jclass clss, jint loc_id, jstring name, jlong size_hint)
{
    const char *grpName = NULL;
    hid_t       group_id = H5I_INVALID_HID;

    UNUSED(clss);

    if (NULL == name)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Gcreate: group name is NULL");

    PIN_JAVA_STRING(ENVONLY, name, grpName, NULL, "H5Gcreate: group name not pinned");

    if ((group_id = H5Gcreate2((hid_t)loc_id, grpName, H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT)) < 0)
        H5_LIBRARY_ERROR(ENVONLY);

done:
    if (grpName)
        UNPIN_JAVA_STRING(ENVONLY, name, grpName);

    return (jint)group_id;
} /* end Java_hdf_hdf5lib_H5__1H5Gcreate */

/*
 * Class:     hdf_hdf5lib_H5
 * Method:    _H5Gcreate2
 * Signature: (ILjava/lang/String;III)I
 */
JNIEXPORT jint JNICALL
Java_hdf_hdf5lib_H5__1H5Gcreate2
    (JNIEnv *env, jclass clss, jint loc_id, jstring name,
          jint link_plist_id, jint create_plist_id, jint access_plist_id)
{
    const char *grpName = NULL;
    hid_t       group_id = H5I_INVALID_HID;

    UNUSED(clss);

    if (NULL == name)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Gcreate2: group name is NULL");

    PIN_JAVA_STRING(ENVONLY, name, grpName, NULL, "H5Gcreate2: group name not pinned");

    if ((group_id = H5Gcreate2((hid_t)loc_id, grpName, (hid_t)link_plist_id, (hid_t)create_plist_id, (hid_t)access_plist_id)) < 0)
        H5_LIBRARY_ERROR(ENVONLY);

done:
    if (grpName)
        UNPIN_JAVA_STRING(ENVONLY, name, grpName);

    return (jint)group_id;
} /* end Java_hdf_hdf5lib_H5__1H5Gcreate2 */

/*
 * Class:     hdf_hdf5lib_H5
 * Method:    _H5Gcreate_anon
 * Signature: (III)I
 */
JNIEXPORT jint JNICALL
Java_hdf_hdf5lib_H5__1H5Gcreate_1anon
    (JNIEnv *env, jclass clss, jint loc_id, jint gcpl_id, jint gapl_id)
{
    hid_t group_id = H5I_INVALID_HID;

    UNUSED(clss);

    if ((group_id = H5Gcreate_anon((hid_t)loc_id, (hid_t)gcpl_id, (hid_t)gapl_id)) < 0)
        H5_LIBRARY_ERROR(ENVONLY);

done:
    return (jint)group_id;
} /* end Java_hdf_hdf5lib_H5__1H5Gcreate_1anon */

/*
 * Class:     hdf_hdf5lib_H5
 * Method:    _H5Gopen
 * Signature: (ILjava/lang/String;)I
 */
JNIEXPORT jint JNICALL Java_hdf_hdf5lib_H5__1H5Gopen
    (JNIEnv *env, jclass clss, jint loc_id, jstring name)
{
    const char *grpName = NULL;
    hid_t       group_id = H5I_INVALID_HID;

    UNUSED(clss);

    if (NULL == name)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Gopen: group name is NULL");

    PIN_JAVA_STRING(ENVONLY, name, grpName, NULL, "H5Gopen: group name not pinned");

    if ((group_id = H5Gopen2((hid_t)loc_id, grpName, H5P_DEFAULT)) < 0)
        H5_LIBRARY_ERROR(ENVONLY);

done:
    if (grpName)
        UNPIN_JAVA_STRING(ENVONLY, name, grpName);

    return (jint)group_id;
} /* end Java_hdf_hdf5lib_H5__1H5Gopen */

/*
 * Class:     hdf_hdf5lib_H5
 * Method:    _H5Gopen2
 * Signature: (ILjava/lang/String;I)I
 */
JNIEXPORT jint JNICALL
Java_hdf_hdf5lib_H5__1H5Gopen2
    (JNIEnv *env, jclass clss, jint loc_id, jstring name, jint access_plist_id)
{
    const char *grpName = NULL;
    hid_t       group_id = H5I_INVALID_HID;

    UNUSED(clss);

    if (NULL == name)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Gopen2: group name is NULL");

    PIN_JAVA_STRING(ENVONLY, name, grpName, NULL, "H5Gopen2: group name not pinned");

    if ((group_id = H5Gopen2((hid_t)loc_id, grpName, (hid_t)access_plist_id)) < 0)
        H5_LIBRARY_ERROR(ENVONLY);

done:
    if (grpName)
        UNPIN_JAVA_STRING(ENVONLY, name, grpName);

    return (jint)group_id;
} /* end Java_hdf_hdf5lib_H5__1H5Gopen2 */


/*
 * Class:     hdf_hdf5lib_H5
 * Method:    H5Gget_create_plist
 * Signature: (I)I
 */
JNIEXPORT jint JNICALL
Java_hdf_hdf5lib_H5_H5Gget_1create_1plist
    (JNIEnv *env, jclass clss, jint loc_id)
{
    hid_t plist_id = H5I_INVALID_HID;

    UNUSED(clss);

    if ((plist_id = H5Gget_create_plist((hid_t)loc_id)) < 0)
        H5_LIBRARY_ERROR(ENVONLY);

done:
    return (jint)plist_id;
} /* end Java_hdf_hdf5lib_H5_H5Gget_1create_1plist */

/*
 * Class:     hdf_hdf5lib_H5
 * Method:    H5Gget_info
 * Signature: (I)Lhdf/hdf5lib/structs/H5G_info_t;
 */
JNIEXPORT jobject JNICALL
Java_hdf_hdf5lib_H5_H5Gget_1info
    (JNIEnv *env, jclass clss, jint loc_id)
{
    H5G_info_t group_info;
    jobject    obj = NULL;

    UNUSED(clss);

    if (H5Gget_info((hid_t)loc_id, &group_info) < 0)
        H5_LIBRARY_ERROR(ENVONLY);

    if (NULL == (obj = create_H5G_info_t(env, group_info)))
        H5_JNI_FATAL_ERROR(ENVONLY, "H5Gget_info: unable to create H5G_info_t object");

done:
    return obj;
} /* end Java_hdf_hdf5lib_H5_H5Gget_1info */

/*
 * Class:     hdf_hdf5lib_H5
 * Method:    H5Gget_info_by_name
 * Signature: (ILjava/lang/String;I)Lhdf/hdf5lib/structs/H5G_info_t;
 */
JNIEXPORT jobject JNICALL
Java_hdf_hdf5lib_H5_H5Gget_1info_1by_1name
    (JNIEnv *env, jclass clss, jint loc_id, jstring name, jint lapl_id)
{
    H5G_info_t  group_info;
    const char *grpName = NULL;
    jobject     obj = NULL;
    herr_t      ret_val = FAIL;

    UNUSED(clss);

    if (NULL == name)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Gget_info_by_name: group name is NULL");

    PIN_JAVA_STRING(ENVONLY, name, grpName, NULL, "H5Gget_info_by_name: group name not pinned");

    if ((ret_val = H5Gget_info_by_name((hid_t)loc_id, grpName, &group_info, (hid_t)lapl_id)) < 0)
        H5_LIBRARY_ERROR(ENVONLY);

    if (NULL == (obj = create_H5G_info_t(env, group_info))) {
        CHECK_JNI_EXCEPTION(ENVONLY, JNI_TRUE);
        H5_JNI_FATAL_ERROR(ENVONLY, "H5Gget_info_by_name: unable to create H5G_info_t object");
    }

done:
    if (grpName)
        UNPIN_JAVA_STRING(ENVONLY, name, grpName);

    return obj;
} /* end Java_hdf_hdf5lib_H5_H5Gget_1info_1by_1name */

/*
 * Class:     hdf_hdf5lib_H5
 * Method:    H5Gget_info_by_idx
 * Signature: (ILjava/lang/String;IIJI)Lhdf/hdf5lib/structs/H5G_info_t;
 */
JNIEXPORT jobject JNICALL
Java_hdf_hdf5lib_H5_H5Gget_1info_1by_1idx
    (JNIEnv *env, jclass clss, jint loc_id, jstring name, jint index_type,
          jint order, jlong n, jint lapl_id)
{
    H5_iter_order_t  corder = (H5_iter_order_t)order;
    H5_index_t       cindex_type = (H5_index_t)index_type;
    H5G_info_t       group_info;
    const char      *grpName = NULL;
    jobject          obj = NULL;
    herr_t           ret_val = FAIL;

    UNUSED(clss);

    if (NULL == name)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Gget_info_by_idx: group name is NULL");

    PIN_JAVA_STRING(ENVONLY, name, grpName, NULL, "H5Gget_info_by_idx: group name not pinned");

    if ((ret_val = H5Gget_info_by_idx((hid_t)loc_id, grpName, cindex_type,
            corder, (hsize_t)n, &group_info, (hid_t)lapl_id)) < 0)
        H5_LIBRARY_ERROR(ENVONLY);

    if (NULL == (obj = create_H5G_info_t(env, group_info))) {
        CHECK_JNI_EXCEPTION(ENVONLY, JNI_TRUE);
        H5_JNI_FATAL_ERROR(ENVONLY, "H5Gget_info_by_idx: unable to create H5G_info_t object");
    }

done:
    if (grpName)
        UNPIN_JAVA_STRING(ENVONLY, name, grpName);

    return obj;
} /* end Java_hdf_hdf5lib_H5_H5Gget_1info_1by_1idx */

/*
 * Class:     hdf_hdf5lib_H5
 * Method:    H5Glink
 * Signature: (IILjava/lang/String;Ljava/lang/String;)I
 */
JNIEXPORT jint JNICALL
Java_hdf_hdf5lib_H5_H5Glink
    (JNIEnv *env, jclass clss, jint loc_id, jint link_type, jstring cur_name, jstring dst_name)
{
    const char *lCurName = NULL;
    const char *lDstName = NULL;
    herr_t      status = FAIL;

    UNUSED(clss);

    if (NULL == cur_name)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Glink: src name is NULL");
    if (NULL == dst_name)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Glink: dst name is NULL");

    PIN_JAVA_STRING(ENVONLY, cur_name, lCurName, NULL, "H5Glink: src name not pinned");
    PIN_JAVA_STRING(ENVONLY, dst_name, lDstName, NULL, "H5Glink: dest name not pinned");

    if ((status = H5Glink((hid_t)loc_id, (H5G_link_t)link_type, lCurName, lDstName)) < 0)
        H5_LIBRARY_ERROR(ENVONLY);

done:
    if (lDstName)
        UNPIN_JAVA_STRING(ENVONLY, dst_name, lDstName);
    if (lCurName)
        UNPIN_JAVA_STRING(ENVONLY, cur_name, lCurName);

    return (jint)status;
} /* end Java_hdf_hdf5lib_H5_H5Glink */

/*
 * Class:     hdf_hdf5lib_H5
 * Method:    H5Glink
 * Signature: (ILjava/lang/String;IILjava/lang/String;)I
 */
JNIEXPORT jint JNICALL
Java_hdf_hdf5lib_H5_H5Glink2
    (JNIEnv *env, jclass clss, jint cur_loc_id, jstring cur_name, jint link_type, jint dst_loc_id, jstring dst_name)
{
    const char *lCurName = NULL;
    const char *lDstName = NULL;
    herr_t      status = FAIL;

    UNUSED(clss);

    if (NULL == cur_name)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Glink2: src name is NULL");
    if (NULL == dst_name)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Glink2: dst name is NULL");

    PIN_JAVA_STRING(ENVONLY, cur_name, lCurName, NULL, "H5Glink2: src name not pinned");
    PIN_JAVA_STRING(ENVONLY, dst_name, lDstName, NULL, "H5Glink2: dest name not pinned");

    if ((status = H5Glink2((hid_t)cur_loc_id, lCurName, (H5G_link_t)link_type, (hid_t)dst_loc_id, lDstName)) < 0)
        H5_LIBRARY_ERROR(ENVONLY);

done:
    if (lDstName)
        UNPIN_JAVA_STRING(ENVONLY, dst_name, lDstName);
    if (lCurName)
        UNPIN_JAVA_STRING(ENVONLY, cur_name, lCurName);

    return (jint)status;
} /* end Java_hdf_hdf5lib_H5_H5Glink2 */

/*
 * Class:     hdf_hdf5lib_H5
 * Method:    H5Gunlink
 * Signature: (ILjava/lang/String;)I
 */
JNIEXPORT jint JNICALL
Java_hdf_hdf5lib_H5_H5Gunlink
    (JNIEnv *env, jclass clss, jint loc_id, jstring name)
{
    const char *grpName = NULL;
    herr_t      status = FAIL;

    UNUSED(clss);

    if (NULL == name)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Gunlink: group name is NULL");

    PIN_JAVA_STRING(ENVONLY, name, grpName, NULL, "H5Gunlink: group name not pinned");

    if ((status = H5Gunlink((hid_t)loc_id, grpName)) < 0)
        H5_LIBRARY_ERROR(ENVONLY);

done:
    if (grpName)
        UNPIN_JAVA_STRING(ENVONLY, name, grpName);

    return (jint)status;
} /* end Java_hdf_hdf5lib_H5_H5Gunlink */

/*
 * Class:     hdf_hdf5lib_H5
 * Method:    H5Gmove
 * Signature: (ILjava/lang/String;Ljava/lang/String;)I
 */
JNIEXPORT jint JNICALL
Java_hdf_hdf5lib_H5_H5Gmove
    (JNIEnv *env, jclass clss, jint loc_id, jstring src_name, jstring dst_name)
{
    const char *lSrcName = NULL;
    const char *lDstName = NULL;
    herr_t      status = FAIL;

    UNUSED(clss);

    if (NULL == src_name)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Gmove: src_name is NULL");
    if (NULL == dst_name)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Gmove: dst_name is NULL");

    PIN_JAVA_STRING(ENVONLY, src_name, lSrcName, NULL, "H5Gmove: src_name not pinned");
    PIN_JAVA_STRING(ENVONLY, dst_name, lDstName, NULL, "H5Gmove: dst_name not pinned");

    if ((status = H5Gmove((hid_t)loc_id, lSrcName, lDstName)) < 0)
        H5_LIBRARY_ERROR(ENVONLY);

done:
    if (lDstName)
        UNPIN_JAVA_STRING(ENVONLY, dst_name, lDstName);
    if (lSrcName)
        UNPIN_JAVA_STRING(ENVONLY, src_name, lSrcName);

    return (jint)status;
} /* end Java_hdf_hdf5lib_H5_H5Gmove */

/*
 * Class:     hdf_hdf5lib_H5
 * Method:    H5Gget_objinfo
 * Signature: (ILjava/lang/String;Z[J[J[I[J)I
 */
JNIEXPORT jint JNICALL
Java_hdf_hdf5lib_H5_H5Gget_1objinfo
    (JNIEnv *env, jclass clss, jint loc_id, jstring name, jboolean follow_link,
            jlongArray fileno, jlongArray objno, jintArray link_info, jlongArray mtime)
{
    const char *grpName = NULL;
    herr_t      status = FAIL;
    jboolean    isCopy;
    jsize       arrLen;
    jlong      *filenoArray = NULL;
    jlong      *objnoArray = NULL;
    jlong      *mtimeArray = NULL;
    jint       *link_infoArray = NULL;
    hbool_t     follow;
    H5G_stat_t  h5gInfo;

    UNUSED(clss);

    if (follow_link == JNI_TRUE) {
        follow = TRUE;  /*  HDF5 'TRUE' */
    }
    else if (follow_link == JNI_FALSE) {
        follow = FALSE;  /*  HDF5 'FALSE' */
    }
    else {
        H5_BAD_ARGUMENT_ERROR(ENVONLY, "H5Gget_objinfo: follow_link is invalid");
    }

    if (NULL == name)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Gget_objinfo: group name is NULL");

    PIN_JAVA_STRING(ENVONLY, name, grpName, NULL, "H5Gget_objinfo: group name not pinned");

    if (NULL == fileno)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Gget_objinfo: fileno is NULL");

    if ((arrLen = ENVPTR->GetArrayLength(ENVONLY, fileno)) < 0) {
        CHECK_JNI_EXCEPTION(ENVONLY, JNI_TRUE);
        H5_BAD_ARGUMENT_ERROR(ENVONLY, "H5Gget_objinfo: fileno array length < 0");
    }
    if (arrLen < 2)
        H5_BAD_ARGUMENT_ERROR(ENVONLY, "H5Gget_objinfo: fileno input array < 2");

    PIN_LONG_ARRAY(ENVONLY, fileno, filenoArray, &isCopy, "H5Gget_objinfo: fileno not pinned");

    if (NULL == objno)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Gget_objinfo: objno is NULL");

    if ((arrLen = ENVPTR->GetArrayLength(ENVONLY, objno)) < 0) {
        CHECK_JNI_EXCEPTION(ENVONLY, JNI_TRUE);
        H5_BAD_ARGUMENT_ERROR(ENVONLY, "H5Gget_objinfo: objno array length < 0");
    }
    if (arrLen < 2)
        H5_BAD_ARGUMENT_ERROR(ENVONLY, "H5Gget_objinfo: objno input array < 2");

    PIN_LONG_ARRAY(ENVONLY, objno, objnoArray, &isCopy, "H5Gget_objinfo: objno not pinned");

    if (NULL == link_info)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Gget_objinfo: link_info is NULL");

    if ((arrLen = ENVPTR->GetArrayLength(ENVONLY, link_info)) < 0) {
        CHECK_JNI_EXCEPTION(ENVONLY, JNI_TRUE);
        H5_BAD_ARGUMENT_ERROR(ENVONLY, "H5Gget_objinfo: link_info array length < 0");
    }
    if (arrLen < 3)
        H5_BAD_ARGUMENT_ERROR(ENVONLY, "H5Gget_objinfo: link_info input array < 3");

    PIN_INT_ARRAY(ENVONLY, link_info, link_infoArray, &isCopy, "H5Gget_objinfo: link_info not pinned");

    if (NULL == mtime)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Gget_objinfo: mtime is NULL");

    if ((arrLen = ENVPTR->GetArrayLength(ENVONLY, mtime)) < 0) {
        CHECK_JNI_EXCEPTION(ENVONLY, JNI_TRUE);
        H5_BAD_ARGUMENT_ERROR(ENVONLY, "H5Gget_objinfo: mtime array length < 0");
    }
    if (arrLen < 1)
        H5_BAD_ARGUMENT_ERROR(ENVONLY, "H5Gget_objinfo: mtime input array < 1");

    PIN_LONG_ARRAY(ENVONLY, mtime, mtimeArray, &isCopy, "H5Gget_objinfo: mtime not pinned");

    if ((status = H5Gget_objinfo((hid_t)loc_id, grpName, follow, &h5gInfo)) < 0)
        H5_LIBRARY_ERROR(ENVONLY);

    filenoArray[0] = (jlong)h5gInfo.fileno[0];
    filenoArray[1] = (jlong)h5gInfo.fileno[1];
    objnoArray[0] = (jlong)h5gInfo.objno[0];
    objnoArray[1] = (jlong)h5gInfo.objno[1];
    mtimeArray[0] = (jlong)h5gInfo.mtime;
    link_infoArray[0] = (jint)h5gInfo.nlink;
    link_infoArray[1] = (jint)h5gInfo.type;
    link_infoArray[2] = (jint)h5gInfo.linklen;

done:
    if (grpName)
        UNPIN_JAVA_STRING(ENVONLY, name, grpName);
    if (mtimeArray)
        UNPIN_LONG_ARRAY(ENVONLY, mtime, mtimeArray, (status < 0) ? JNI_ABORT : 0);
    if (link_infoArray)
        UNPIN_INT_ARRAY(ENVONLY, link_info, link_infoArray, (status < 0) ? JNI_ABORT : 0);
    if (objnoArray)
        UNPIN_LONG_ARRAY(ENVONLY, objno, objnoArray, (status < 0) ? JNI_ABORT : 0);
    if (filenoArray)
        UNPIN_LONG_ARRAY(ENVONLY, fileno, filenoArray, (status < 0) ? JNI_ABORT : 0);

    return (jint)status;
}

/*
 * Class:     hdf_hdf5lib_H5
 * Method:    H5Gget_linkval
 * Signature: (ILjava/lang/String;I[Ljava/lang/String;)I
 */
JNIEXPORT jint JNICALL
Java_hdf_hdf5lib_H5_H5Gget_1linkval
    (JNIEnv *env, jclass clss, jint loc_id, jstring name, jint size, jobjectArray value)
{
    const char *grpName = NULL;
    herr_t      status = FAIL;
    char       *lValue = NULL;
    jsize       arrLen;
    jstring     jstr;

    UNUSED(clss);

    if (NULL == value)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Gget_linkval: value buffer is NULL");

    if ((arrLen = ENVPTR->GetArrayLength(ENVONLY, value)) <= 0) {
        CHECK_JNI_EXCEPTION(ENVONLY, JNI_TRUE);
        H5_BAD_ARGUMENT_ERROR(ENVONLY, "H5Gget_linkval: value buffer length <= 0");
    }

    if (size < 0)
        H5_BAD_ARGUMENT_ERROR(ENVONLY, "H5Gget_linkval: size < 0");

    /* allow for NULL terminator */
    if (NULL == (lValue = (char *) HDcalloc((size_t)size + 1, sizeof(char))))
        H5_JNI_FATAL_ERROR(ENVONLY, "H5Gget_linkval: malloc failed");

    if (NULL == name)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Gget_linkval: name is NULL");

    PIN_JAVA_STRING(ENVONLY, name, grpName, NULL, "H5Gget_linkval: name not pinned");

    if ((status = H5Gget_linkval((hid_t)loc_id, grpName, (size_t)size + 1, lValue)) < 0)
        H5_LIBRARY_ERROR(ENVONLY);
    lValue[size] = '\0';

    if (NULL == (jstr = ENVPTR->NewStringUTF(ENVONLY, lValue))) {
        CHECK_JNI_EXCEPTION(ENVONLY, JNI_TRUE);
        H5_JNI_FATAL_ERROR(ENVONLY, "H5Gget_linkval: out of memory - unable to construct string from UTF characters");
    }

    ENVPTR->SetObjectArrayElement(ENVONLY, value, 0, jstr);
    CHECK_JNI_EXCEPTION(ENVONLY, JNI_FALSE);

    ENVPTR->DeleteLocalRef(ENVONLY, jstr);

done:
    if (grpName)
        UNPIN_JAVA_STRING(ENVONLY, name, grpName);
    if (lValue)
        HDfree(lValue);

    return (jint)status;
}

/*
 * Class:     hdf_hdf5lib_H5
 * Method:    H5Gset_comment
 * Signature: (ILjava/lang/String;Ljava/lang/String;)I
 */
JNIEXPORT jint JNICALL
Java_hdf_hdf5lib_H5_H5Gset_1comment
    (JNIEnv *env, jclass clss, jint loc_id, jstring name, jstring comment)
{
    const char *lGrpName = NULL;
    const char *lCommentName = NULL;
    herr_t      status = FAIL;

    UNUSED(clss);

    if (NULL == name)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Gset_comment: name is NULL");
    if (NULL == comment)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Gset_comment: comment is NULL");

    PIN_JAVA_STRING(ENVONLY, name, lGrpName, NULL, "H5Gset_comment: name not pinned");
    PIN_JAVA_STRING(ENVONLY, comment, lCommentName, NULL, "H5Gset_comment: comment not pinned");

    if ((status = H5Gset_comment((hid_t)loc_id, lGrpName, lCommentName)) < 0)
        H5_LIBRARY_ERROR(ENVONLY);

done:
    if (lCommentName)
        UNPIN_JAVA_STRING(ENVONLY, comment, lCommentName);
    if (lGrpName)
        UNPIN_JAVA_STRING(ENVONLY, name, lGrpName);

    return (jint)status;
}

/*
 * Class:     hdf_hdf5lib_H5
 * Method:    H5Gget_comment
 * Signature: (ILjava/lang/String;I[Ljava/lang/String;)I
 */
JNIEXPORT jint JNICALL
Java_hdf_hdf5lib_H5_H5Gget_1comment
    (JNIEnv *env, jclass clss, jint loc_id, jstring name, jint bufsize, jobjectArray comment)
{
    const char *grpName = NULL;
    int         comment_size = FAIL;
    char       *lComment = NULL;
    jsize       arrLen;
    jstring     jstr;

    UNUSED(clss);

    if (NULL == comment)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Gget_comment: comment buffer is NULL");

    if ((arrLen = ENVPTR->GetArrayLength(ENVONLY, comment)) <= 0) {
        CHECK_JNI_EXCEPTION(ENVONLY, JNI_TRUE);
        H5_BAD_ARGUMENT_ERROR(ENVONLY, "H5Gget_comment: comment buffer length <= 0");
    }

    if (bufsize < 0)
        H5_BAD_ARGUMENT_ERROR(ENVONLY, "H5Gget_comment: bufsize < 0");

    /* allow for NULL terminator */
    if (NULL == (lComment = (char *) HDcalloc((size_t)bufsize + 1, sizeof(char))))
        H5_JNI_FATAL_ERROR(ENVONLY, "H5Gget_comment: malloc failed");

    if (NULL == name)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Gget_comment: name is NULL");

    PIN_JAVA_STRING(ENVONLY, name, grpName, NULL, "H5Gget_comment: name not pinned");

    if ((comment_size = H5Gget_comment((hid_t)loc_id, grpName, (size_t)bufsize + 1, lComment)) < 0)
        H5_LIBRARY_ERROR(ENVONLY);
    lComment[bufsize] = '\0';

    if (NULL == (jstr = ENVPTR->NewStringUTF(ENVONLY, lComment))) {
        CHECK_JNI_EXCEPTION(ENVONLY, JNI_TRUE);
        H5_JNI_FATAL_ERROR(ENVONLY, "H5Gget_comment: out of memory - unable to construct string from UTF characters");
    }

    ENVPTR->SetObjectArrayElement(ENVONLY, comment, 0, jstr);
    CHECK_JNI_EXCEPTION(ENVONLY, JNI_FALSE);

    ENVPTR->DeleteLocalRef(ENVONLY, jstr);

done:
    if (grpName)
        UNPIN_JAVA_STRING(ENVONLY, name, grpName);
    if (lComment)
        HDfree(lComment);

    return (jint)comment_size;
}

/*
 * Class:     hdf_hdf5lib_H5
 * Method:    H5Gget_num_objs
 * Signature: (I[J)I
 */
JNIEXPORT jint JNICALL
Java_hdf_hdf5lib_H5_H5Gget_1num_1objs
    (JNIEnv *env, jclass clss, jint loc_id, jlongArray num_obj)
{
    jboolean  isCopy;
    jlong    *num_objP = NULL;
    hsize_t  *num_objA = NULL;
    size_t    i;
    jsize     rank;
    herr_t    status = FAIL;

    UNUSED(clss);

    if (NULL == num_obj)
        H5_NULL_ARGUMENT_ERROR(ENVONLY, "H5Gget_num_objs: num_obj is NULL");

    PIN_LONG_ARRAY(ENVONLY, num_obj, num_objP, &isCopy, "H5Gget_num_objs: num_obj not pinned");

    if ((rank = (size_t)ENVPTR->GetArrayLength(ENVONLY, num_obj)) <= 0) {
        CHECK_JNI_EXCEPTION(ENVONLY, JNI_TRUE);
        H5_BAD_ARGUMENT_ERROR(ENVONLY, "H5Gget_num_objs: num_obj buffer length <= 0");
    }

    if (NULL == (num_objA = (hsize_t *)HDmalloc(rank * sizeof(hsize_t))))
        H5_JNI_FATAL_ERROR(ENVONLY, "H5Fget_obj_ids: malloc failed");

    if ((status = H5Gget_num_objs((hid_t)loc_id, (hsize_t *)num_objA)) < 0)
        H5_LIBRARY_ERROR(ENVONLY);

    for (i = 0; i < (size_t)rank; i++) {
        num_objP[i] = (jlong)num_objA[i];
    } /* end for */

done:
    if (num_objA)
        HDfree(num_objA);
    if (num_objP)
        UNPIN_LONG_ARRAY(ENVONLY, num_obj, num_objP, (status < 0) ? JNI_ABORT : 0);

    return (jint)status;
}

/*
 * Class:     hdf_hdf5lib_H5
 * Method:    H5Gget_objname_by_idx
 * Signature: (IJ[Ljava/lang/String;J)J
 */
JNIEXPORT jlong JNICALL
Java_hdf_hdf5lib_H5_H5Gget_1objname_1by_1idx
    (JNIEnv *env, jclass clss, jint loc_id, jlong idx, jobjectArray obj_name, jlong size)
{
    herr_t      status = FAIL;
    char       *lObjName = NULL;
    jsize       arrLen;
    jstring     jstr;
    ssize_t     length;

    UNUSED(clss);

    if (obj_name != NULL) {
        if (size <= 0)
            H5_BAD_ARGUMENT_ERROR(ENVONLY, "H5Gget_objname_by_idx: size <= 0");

        /* allow for NULL terminator */
        if (NULL == (lObjName = (char *) HDcalloc((size_t)size + 1, sizeof(char))))
            H5_JNI_FATAL_ERROR(ENVONLY, "H5Gget_objname_by_idx: malloc failed");

        if ((length = H5Gget_objname_by_idx((hid_t)loc_id, (hsize_t)idx, lObjName, (size_t)(size + 1))) < 0)
            H5_LIBRARY_ERROR(ENVONLY);

        lObjName[size] = '\0';
        if (NULL == (jstr = ENVPTR->NewStringUTF(ENVONLY, lObjName))) {
            CHECK_JNI_EXCEPTION(ENVONLY, JNI_TRUE);
            H5_JNI_FATAL_ERROR(ENVONLY, "H5Gget_objname_by_idx: out of memory - unable to construct string from UTF characters");
        }

        ENVPTR->SetObjectArrayElement(ENVONLY, obj_name, 0, jstr);
        CHECK_JNI_EXCEPTION(ENVONLY, JNI_FALSE);

        ENVPTR->DeleteLocalRef(ENVONLY, jstr);
    }
    else {
        if ((length = H5Gget_objname_by_idx((hid_t)loc_id, (hsize_t)idx, NULL, 0)) < 0)
            H5_LIBRARY_ERROR(ENVONLY);
    }
done:
    if (lObjName)
        HDfree(lObjName);

    return (jlong)length;
}


/*
 * Class:     hdf_hdf5lib_H5
 * Method:    H5Gget_objtype_by_idx
 * Signature: (IJ)I
 */
JNIEXPORT jint JNICALL
Java_hdf_hdf5lib_H5_H5Gget_1objtype_1by_1idx
    (JNIEnv *env, jclass clss, jint group_id, jlong idx)
{
    int retVal = -1;

    UNUSED(clss);

    if ((retVal = H5Gget_objtype_by_idx((hid_t)group_id, (hsize_t)idx)) < 0)
        H5_LIBRARY_ERROR(ENVONLY);

done:
    return (jint)retVal;
}

#ifdef __cplusplus
} /* end extern "C" */
#endif /* __cplusplus */
